# Copyright 2013 Tom SF Haines

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import os
import os.path
import time
import gzip
import shutil
import json
import numpy

from gi.repository import Gtk

from viewer import Viewer
from tile_image import TileImage

from depth_sensor import DepthSensor, c_float_p, to_coord
import ctypes



c_ubyte_p = ctypes.POINTER(ctypes.c_ubyte)
c_int_p = ctypes.POINTER(ctypes.c_int)



class DepthRecorder(Gtk.Window):
  TIME_MODE_DRIVER_MICRO = 0 # What it should output
  TIME_MODE_DRIVER_MILLI = 1 # *1000
  TIME_MODE_DRIVER_NANO = 2 #  /1000
  TIME_MODE_SYSTEM = 3 # Use the system clock - not as accurate, but will work.
  TIME_MODE_SEQUENTIAL = 4 # One after the other, time be dammed - stupid.
  
  def __init__(self):
    # Do the basic window setup...
    Gtk.Window.__init__(self, title='Depth Recorder')
    self.connect('delete-event', Gtk.main_quit)
    self.set_default_size(1024, 576)
    
    # Setup a viewer with a single image...
    self.viewer = Viewer()
    self.viewer.set_bg(1.0, 1.0, 1.0)
    
    self.image = TileImage(None)
    self.viewer.add_layer(self.image)
    
    self.viewer.set_on_draw(self.__on_draw)
    
    # Setup the menu system...
    self.uimanager = self.__create_uimanager()
    self.add_accel_group(self.uimanager.get_accel_group())
    menu_bar = self.uimanager.get_widget('/menu_bar')
    icon_bar = self.uimanager.get_widget('/icon_bar')
    
    # Layout the window...
    vertical = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
    vertical.pack_start(menu_bar, False, False, 0)
    vertical.pack_start(self.viewer, True, True, 0)
    vertical.pack_start(icon_bar, False, False, 0)
    self.add(vertical)
    
    # Setup the depth sensor, set it going...
    self.ds = DepthSensor()
    self.ds.start()
    self.norm = 1.0
    
    # Details for recording...
    self.record = False
    self.output_dir = 'output'
    self.start_time = None
    self.frame = -1
    self.fps = 24.0
    
    self.time_mode = self.TIME_MODE_DRIVER_MICRO
    self.display_record = True
    self.rotate_sane = True
    self.two_step = True
    
    self.index = None # Temporary intermediate used when recording.
    self.temp = None # Temporary buffer used when recording.
    
    # Load the configuration, if it exists...
    self.load_config()


  def __del__(self):
    self.ds.stop()
  
  
  def __create_uimanager(self):
    # Create the actions for the file menu action group...
    action_file = Gtk.Action('file', 'File', None, None)
    
    action_quit = Gtk.Action('quit', 'Quit', 'Terminates this application', Gtk.STOCK_QUIT)
    action_quit.connect('activate', Gtk.main_quit)
    
    group_file = Gtk.ActionGroup('File')
    group_file.add_action_with_accel(action_file, '')
    group_file.add_action_with_accel(action_quit, '<ctrl>q')
    
    
    # Create the actions for the operations menu action group...
    action_ops = Gtk.Action('ops', 'Operations', None, None)
    
    self.action_record = Gtk.ToggleAction('record', 'Record','Toggles recording a depth-video', Gtk.STOCK_MEDIA_RECORD)
    self.action_record.connect('activate', self.__record)
    
    self.action_configure = Gtk.Action('configure', 'Configure...', 'Configures the application', Gtk.STOCK_PREFERENCES)
    self.action_configure.connect('activate', self.__configure)
    
    group_ops = Gtk.ActionGroup('Operations')
    group_ops.add_action_with_accel(action_ops, '')
    group_ops.add_action_with_accel(self.action_record, '<space>')
    group_ops.add_action_with_accel(self.action_configure, '<ctrl>c')
    
    
    # The ui configuration string...
    ui = """
    <ui>
     <menubar name='menu_bar'>
      <menu action='file'>
       <menuitem action='quit'/>
      </menu>
      <menu action='ops'>
       <menuitem action='record'/>
       <menuitem action='configure'/>
      </menu>
     </menubar>
     <toolbar name='icon_bar'>
      <toolitem action='record'/>
      <toolitem action='configure'/>
     </toolbar>
    </ui>
    """
    
    
    # Create and return the actual ui manager...
    ret = Gtk.UIManager()
    ret.add_ui_from_string(ui)
    ret.insert_action_group(group_file)
    ret.insert_action_group(group_ops)
    
    return ret
  
  
  def __record(self, widget):
    self.record = widget.get_active()
    
    if self.record:
      self.start_time = None
      self.frame = 0
      try:
        os.makedirs(self.output_dir)
      except: pass
    else:
      # If needed do the conversion in post...
      if self.two_step:
        # Get all the files with a .npy extension in the output directory...
        to_proc = filter(lambda fn: fn[-4:]=='.npy', os.listdir(self.output_dir))
        
        # Setup a progress bar...
        dialog = Gtk.Dialog('Converting to water simulation')
        pb = Gtk.ProgressBar()
        dialog.get_content_area().pack_start(pb, False, False, 0)
        dialog.show_all()
        
        # Iterate and process each npy file in turn...
        self.draw = False
        
        for i, fn in enumerate(to_proc):
          pb.set_fraction(float(i) / len(to_proc))
          print 'Processing %s'%fn
          while Gtk.events_pending():
            Gtk.main_iteration()
          
          # Load the array...
          cfn = os.path.join(self.output_dir, fn)
          f = open(cfn, 'rb')
          frame = dict()
          frame['data'] = numpy.load(f)
          f.close()
          
          # Do the work...
          self.frame = int(fn[:-4])
          self.save_frame(frame)
          
          # Clean up the intermediate file...
          os.unlink(cfn)
        
        self.draw = True
        self.viewer.queue_draw()
        
        # Remove the progress bar...
        dialog.destroy()
      
      # Move to next recording point...
      num = 0
      trim = 0
      for i in xrange(-1, -len(self.output_dir), -1):
        if not self.output_dir[i].isdigit(): break
        trim += 1
        num += (ord(self.output_dir[i]) - ord('0')) * (10 ** (-i - 1))
      
      num += 1
      self.output_dir = self.output_dir[:len(self.output_dir)-trim] + str(num)


  def __on_draw(self):
    # Check if a new frame is avaliable...
    if True:
      # Fetch the frame...
      frame = self.ds.frame()
      grab_time = time.time()
      
      # Make the frame visible...
      if not self.record or self.display_record:
        self.norm = max(self.norm, frame['data'].max())
        
        image = (255.0 * frame['data'][:,::-1] / self.norm).astype(numpy.uint8)
        image = image.reshape((image.shape[0], image.shape[1], 1))
      
        alpha = numpy.empty(image.shape, dtype=numpy.uint8)
        alpha[:,:,:] = 255
        image = numpy.concatenate([image, image, image, alpha], axis=2).copy('C')
      
        self.image.from_array(image)
      
      # Record the frame...
      if self.record:
        if self.start_time==None:
          exp_frame = 0
          if self.time_mode==self.TIME_MODE_DRIVER_MICRO:
            self.start_time = frame['time']
          elif self.time_mode==self.TIME_MODE_DRIVER_MILLI:
            self.start_time = frame['time'] * 1000.0
          elif self.time_mode==self.TIME_MODE_DRIVER_NANO:
            self.start_time = frame['time'] / 1000.0
          elif self.time_mode==self.TIME_MODE_SYSTEM:
            self.start_time = grab_time
          else: # Sequential
            self.start_time = 0.0
        else:
          if self.time_mode==self.TIME_MODE_DRIVER_MICRO:
            exp_frame = int((frame['time'] - self.start_time) * self.fps)
          elif self.time_mode==self.TIME_MODE_DRIVER_MILLI:
            exp_frame = int((frame['time'] * 1000.0 - self.start_time) * self.fps)
          elif self.time_mode==self.TIME_MODE_DRIVER_NANO:
            exp_frame = int((frame['time'] / 1000.0 - self.start_time) * self.fps)
          elif self.time_mode==self.TIME_MODE_SYSTEM:
            exp_frame = int((grab_time - self.start_time) * self.fps)
          else: # Sequential
            exp_frame = self.frame + 1
          
        
        if exp_frame>self.frame:
          self.frame = exp_frame
          
          if self.two_step:
            fn = os.path.join(self.output_dir, '%04d.npy'%self.frame)
            
            f = open(fn, 'wb')
            numpy.save(f, frame['data'])
            f.close()
            
          else:
            self.save_frame(frame)
    
    # Do it all again!..
    if self.draw:
      self.viewer.queue_draw()


  def save_frame(self, frame):
    # Convert to vertices...
    verts = self.ds.to_pos(frame)
        
    # Resize the indices as needed...
    if self.index==None or self.index.shape[0]!=verts.shape[0] or self.index.shape[1]!=verts.shape[1]:
      self.index = numpy.empty((verts.shape[0], verts.shape[1]), dtype=numpy.int32)
        
    # Resize the buffer as needed...
    max_size = 4 * (4 + 3 * 4 * (verts.shape[0] * verts.shape[1]))
    if self.temp==None or self.temp.shape[0]<max_size:
      self.temp = numpy.empty(max_size, dtype=numpy.uint8)
        
    # Fill the buffer with the uncompressed data...
    size = to_coord.make_bobj(verts.shape[1], verts.shape[0], verts.ctypes.data_as(c_float_p), self.index.ctypes.data_as(c_int_p), self.temp.ctypes.data_as(c_ubyte_p), 1 if self.rotate_sane else 0)
        
    # Splat it into the actual output file...
    fn = os.path.join(self.output_dir, 'fluidsurface_final_%04d.bobj.gz'%self.frame)
        
    f = gzip.open(fn, 'wb', 1)
    f.write(self.temp[:size].tostring())
    f.close()
        
    pfn = os.path.join(self.output_dir, 'fluidsurface_preview_%04d.bobj.gz'%self.frame)
    shutil.copyfile(fn, pfn)
  
  
  def load_config(self):
    try:
      f = open('config.json', 'r')
      
      config = json.loads(f.read())
      
      if 'output_dir' in config:
        self.output_dir = config['output_dir']

      if 'fps' in config:
        self.fps = float(config['fps'])
      
      if 'time_mode' in config:
        tms = config['time_mode']
        if tms=='driver_micro':
          self.time_mode = self.TIME_MODE_DRIVER_MICRO
        elif tms=='driver_milli':
          self.time_mode = self.TIME_MODE_DRIVER_MILLI
        elif tms=='driver_nano':
          self.time_mode = self.TIME_MODE_DRIVER_NANO
        elif tms=='system':
          self.time_mode = self.TIME_MODE_SYSTEM
        elif tms=='sequential':
          self.time_mode = self.TIME_MODE_SEQUENTIAL
      
      if 'display_record' in config:
        self.display_record = bool(config['display_record'])
      
      if 'rotate_sane' in config:
        self.rotate_sane = bool(config['rotate_sane'])
        
      if 'two_step' in config:
        self.two_step = bool(config['two_step'])
    
      f.close()
    except IOError:
      print 'No configuration found - using defaults'
      
  
  def save_config(self):
    config = dict()
    
    config['output_dir'] = self.output_dir
    config['fps'] = self.fps
    
    if self.time_mode==self.TIME_MODE_DRIVER_MICRO:
      config['time_mode'] = 'driver_micro'
    elif self.time_mode==self.TIME_MODE_DRIVER_MILLI:
      config['time_mode'] = 'driver_milli'      
    elif self.time_mode==self.TIME_MODE_DRIVER_NANO:
      config['time_mode'] = 'driver_nano'
    elif self.time_mode==self.TIME_MODE_SYSTEM:
      config['time_mode'] = 'system'
    elif self.time_mode==self.TIME_MODE_SEQUENTIAL:
      config['time_mode'] = 'sequential'
          
    config['display_record'] = self.display_record
    config['rotate_sane'] = self.rotate_sane
    config['two_step'] = self.two_step
    
    f = open('config.json', 'w')
    f.write(json.dumps(config))
    f.close()
  
  
  def __configure(self, widget):
    # Basic dialog setup...
    dialog = Gtk.Dialog('Thresholding Parameters')
    
    dialog.add_button('Cancel', Gtk.ResponseType.CANCEL)
    dialog.add_button('Accept', Gtk.ResponseType.OK)
    dialog.add_button('Accept & Save', Gtk.ResponseType.APPLY)
    
    # Create the various inputs...
    output_dir = Gtk.Entry()
    output_dir.set_text(self.output_dir)
    
    fps = Gtk.SpinButton()
    fps.set_adjustment(Gtk.Adjustment(self.fps, 1.0, 60.0, 0.1, 1.0, 1.0))
    fps.set_digits(1)
    
    time_mode = Gtk.ComboBoxText()
    time_mode.append_text('Driver - micro seconds')
    time_mode.append_text('Driver - milli seconds')
    time_mode.append_text('Driver - nano seconds')
    time_mode.append_text('System')
    time_mode.append_text('Sequential')
    time_mode.set_active(self.time_mode)
    
    display_record = Gtk.CheckButton('Visualise whilst Recording')
    display_record.set_active(self.display_record)
    
    rotate_sane = Gtk.CheckButton('Rotate to Normal Blender Orientation')
    rotate_sane.set_active(self.rotate_sane)
    
    two_step = Gtk.CheckButton('Process after Recording')
    two_step.set_active(self.two_step)
    
    # Stuff them into the dialog...
    box = dialog.get_content_area()
    
    l = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
    l.pack_start(Gtk.Label('Time Source:'), False, False, 8)
    l.pack_start(time_mode, True, True, 0)
    box.pack_start(l, False, False, 0)
    
    l = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
    l.pack_start(Gtk.Label('Output Directory:'), False, False, 8)
    l.pack_start(output_dir, True, True, 0)
    box.pack_start(l, False, False, 0)
    
    l = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
    l.pack_start(Gtk.Label('Frames per Second:'), False, False, 8)
    l.pack_start(fps, True, True, 0)
    box.pack_start(l, False, False, 0)
    
    box.pack_start(display_record, False, False, 0)
    box.pack_start(rotate_sane, False, False, 0)
    box.pack_start(two_step, False, False, 0)
    
    box.show_all()
    
    # Run the dialog...
    response = dialog.run()
    
    # React to the response...
    if response==Gtk.ResponseType.OK or response==Gtk.ResponseType.APPLY:
      self.output_dir = output_dir.get_text()
      self.fps = fps.get_value()
      self.time_mode = time_mode.get_active()
      self.display_record = display_record.get_active()
      self.rotate_sane = rotate_sane.get_active()
      self.two_step = two_step.get_active()
    
    if response==Gtk.ResponseType.APPLY:
      self.save_config()
    
    # Clean up...
    dialog.destroy()