# Copyright 2013 Tom SF Haines

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import ctypes
import numpy
import numpy.ctypeslib



openni = ctypes.CDLL('libOpenNI2.so')
to_coord = ctypes.CDLL('to_coord.so')



class VideoMode(ctypes.Structure):
  _fields_ = [('format', ctypes.c_int),
              ('width', ctypes.c_int),
              ('height', ctypes.c_int),
              ('fps', ctypes.c_int),]


        
class SensorInfo(ctypes.Structure):
  _fields_ = [('stype', ctypes.c_int),
              ('count', ctypes.c_int),
              ('modes', ctypes.POINTER(VideoMode))]


         
class Frame(ctypes.Structure):
  _fields_ = [('data_size', ctypes.c_int),
              ('data', ctypes.POINTER(ctypes.c_ushort)),
              ('stype', ctypes.c_int),
              ('time', ctypes.c_ulonglong),
              ('index', ctypes.c_int),
              ('width', ctypes.c_int),
              ('height', ctypes.c_int),
              ('mode', ctypes.c_int),
              ('crop', ctypes.c_int),
              ('crop_x', ctypes.c_int),
              ('crop_y', ctypes.c_int),
              ('stride', ctypes.c_int)]


c_short_p = ctypes.POINTER(ctypes.c_short)
c_float_p = ctypes.POINTER(ctypes.c_float)



class DepthSensor:
  def __init__(self):
    if openni.oniInitialize(202000030)!=0:
      raise RuntimeError('Failed to initalise OpenNI')
    
    self.device = ctypes.c_void_p()
    if openni.oniDeviceOpen(None, ctypes.pointer(self.device))!=0:
      del self.device
      raise RuntimeError('Failed to connect to device')
    
    self.stream = ctypes.c_void_p()
    if openni.oniDeviceCreateStream(self.device, 3, ctypes.pointer(self.stream))!=0:
      del self.stream
      raise RuntimeError('Failed to open depth stream')
  
  
  def __del__(self):
    if hasattr(self, 'stream'):
      openni.oniStreamDestroy(self.stream)
    
    if hasattr(self, 'device'):
      openni.oniDeviceClose(self.device)
    
    openni.oniShutdown()
  
  
  def start(self):
    if openni.oniStreamStart(self.stream)!=0:
      raise RuntimeError('Could not start depth stream')
  
  
  def stop(self):
    openni.oniStreamStop(self.stream)
 
 
  def frame(self):
    """Grabs a frame, blocking if need be. returns a dictionary of properties, including 'data', a numpy array of depth values."""
    
    # Get the data...
    frame = ctypes.POINTER(Frame)()
    if openni.oniStreamReadFrame(self.stream, ctypes.pointer(frame))!=0:
      raise RuntimeError('Frame grab failed')
      
    # Create the return object...
    ret = dict()
    
    ret['mode'] = frame.contents.mode
    ret['time'] = frame.contents.time / 1000000.0
    ret['data'] = numpy.ctypeslib.as_array(frame.contents.data, (frame.contents.height,frame.contents.width))
    
    # Release the data...
    openni.oniFrameRelease(frame)
    
    # Return...
    return ret


  def to_pos(self, frame):
    """Given a dictionary returned by the frame() call this returns a new numpy array, [x,y,indices of the 3 ordinates] of the depth converted to position, in meters."""
    data = frame['data']
    ret = numpy.empty((data.shape[0], data.shape[1], 3), dtype=numpy.float32)
    
    f = ctypes.c_void_p.in_dll(openni, 'oniCoordinateConverterDepthToWorld')
    
    status = to_coord.map_to_coord(int(data.shape[1]), int(data.shape[0]), data.ctypes.data_as(c_short_p), ret.ctypes.data_as(c_float_p), self.stream, openni.oniCoordinateConverterDepthToWorld)
    if status!=0: raise RuntimeError('Coordinate conversion failure')
    
    return ret
