# Copyright 2013 Tom SF Haines

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



class Viewport:
  """Defines a view onto a renderable region. You have the width and height of the window into which to render - these are always integers; and the start and end coordinates of a viewport for you to render into that window. This exists to be provided to a draw method with a cairo context. Tradition dictates that the aspect ratio of the window and viewport be identical - a renderer is allowed to operate under that assumption."""
  def __init__(self, width, height, start_x, start_y, end_x, end_y):
    self.width = width
    self.height = height
    
    self.start_x = start_x
    self.start_y = start_y
    self.end_x = end_x
    self.end_y = end_y
  
  def __str__(self):
    return 'Viewport x(%.3f-%.3f) y(%.3f-%.3f)' % (self.start_x,self.end_x,self.start_y,self.end_y)
  
  
  def view_to_original(self, x, y):
    x = (x / float(self.width))  * (self.end_x - self.start_x) + self.start_x
    y = (y / float(self.height)) * (self.end_y - self.start_y) + self.start_y
    return (x, y)
  
  def original_to_view(self, x, y):
    x = self.width  * (x - self.start_x) / (self.end_x - self.start_x)
    y = self.height * (y - self.start_y) / (self.end_y - self.start_y)
    return (x, y)
  
  def original_to_view_scale(self, scale):
    return scale * float(self.width) / (self.end_x - self.start_x)



class Layer:
  """Defines an interface that layers have to impliment - basically this allows them to render to the drawing area, and influence things such as its default size."""
  def get_size(self):
    """Returns the tuple (height, width), the default size of the layer (Origin is assumed to be (0,0)). This basically defines the area in which a Viewport can be expected to show something interesting."""
    raise NotImplementedError
  
  def draw(self, ctx, vp):
    """Draws to the provided cairo context (ctx) the contents defined by the provided viewport (vp). Note that it could be in a stack with other layers - it probably shouldn't be clearing the background."""
    raise NotImplementedError
